(function(root, factory) {
    const ns = "webrtc";
    if (typeof define === "function" && define.amd) {
        // AMD
        define([], factory);
    } else if (typeof exports === "object") {
        // Node, CommonJS-like
        module.exports = factory(ns);
    } else {
        // Browser globals (root is window)
        root[ns] = factory(ns);
    }
})(this, (namespace = "webrtc") => {
    let registry = null;
    const container = document.querySelector("#webRtcOptions");
    const NS = namespace;

    let globalDefaults = null;
    let globalSchema = null;
    let globalReadOnly = null;
    let globalStoreArea = null;

    return function() {
        createWebRtcDomElements();
        setStorageArea();
        setOnChangeEvents();

        const methods = {
            show: showWebRtcDom,
            update: updateDefaultData,
            data: () => ({ ...globalDefaults }),
            list: (key = null) => {
                const obj = key && registry[key] ? registry[key] : registry;
                console.log(obj);
                return { ...obj };
            }
        };

        return methods;
    };

    function setStorageArea() {
        // returns undefined
        //
        // Sets global variable governing the storage area.
        // A "managed" area will clear any "local" persisted values

        chrome.storage.managed.get(NS, data => {
            if (data && data[NS]) {
                console.log("Using MANAGED storage.");
                globalStoreArea = "managed";

                console.warn("Clearing LOCAL data.");
                chrome.storage.local.clear();
                return;
            }

            chrome.storage.local.get(NS, data => {
                console.log("Using LOCAL storage.");
                globalStoreArea = "local";
                if (data && !data[NS]) {
                    // console.warn("No [%s] data exists. Using defaults", NS);
                }
            });
        });
    }

    function setOnChangeEvents() {
        chrome.storage.onChanged.addListener((changes, area) => {
            if (!changes || !changes[NS]) return;
            console.log("[%s] received changes in (%s)", NS, area);
            updateDefaultData();
        });
    }

    function updateDefaultData() {
        let showDom;
        let readOnly;
        let area = globalStoreArea;

        if (!area) {
            console.error("No storage area available");
            return;
        }

        switch (area) {
            case "managed":
                readOnly = true;
                showDom = false;
                break;

            case "local":
                readOnly = false;
                showDom = true;
                break;

            default:
                break;
        }

        chrome.storage[area].get(NS, obj => {
            console.log("Update data with [%s] settings", area.toUpperCase());

            // merge actual data into default data
            globalDefaults = { ...globalDefaults, ...obj };

            // set any area specific flags
            globalReadOnly = readOnly;
            showWebRtcDom(showDom);

            parseDataWithSchema();
        });
    }

    function showWebRtcDom(visible) {
        container.style.display = visible ? "block" : "none";
    }

    function createWebRtcDomElements() {
        // returns undefined
        //
        // loads default data and json-schema
        // and caches a shallow copy as a global variable
        // then renders the section banner (run-once)
        // then constructs Form region as HTMLDivElement(s)
        // then refreshes default data with persisted values
        // then executes onComplete callback

        if (registry) return; //! Only run once

        registry = {};

        const urlDefaults = "managed-defaults.json";
        const urlSchema = "managed-schema.json";

        const fetchDefaults = fetch(chrome.runtime.getURL(urlDefaults)).then(r => r.json());
        const fetchSchema = fetch(chrome.runtime.getURL(urlSchema)).then(r => r.json());

        Promise.all([fetchDefaults, fetchSchema])
            .then(values => {
                globalDefaults = { ...values[0] };
                globalSchema = { ...values[1] };
            })
            .then(renderSectionBanner)
            .then(parseDataWithSchema)
            .then(updateDefaultData)
            .then(onComplete);
    }

    function onComplete() {
        console.log("WebRTC complete. Trigger onComplete()");
    }

    function renderSectionBanner() {
        // returns undefined
        //
        // Inserts the meta data section banner
        // into the DOM container

        let schema = { ...globalSchema };

        // Create the container and the banner description
        const { title, description } = schema.properties[NS];
        const banner = document.createElement("div");
        banner.innerHTML = `
                        <div class="op-heading" ${i18n(title)}></div>
                        <div class="op-description" ${i18n(description)}</div>
                        <div class="opHr"></div>`;

        container.append(banner);
    }

    function renderSeperator(key, path) {
        // return undefined
        //
        // inserts an empty HTMLDivElement for styling purposes

        const div = document.createElement("div");
        const cssname = path.join("-").toLowerCase();
        const zero = /^0$/.test(key);

        div.classList.add("op-object");
        div.classList.add(`op-object-${cssname}`);

        if (zero) div.classList.add("op-object-first");

        container.append(div);
    }

    function parseDataWithSchema() {
        // returns undefined
        //
        // uses cached data to construct DOM form elements
        // and other UI meta Objects (titles, desc etc..)

        let defaults = { ...globalDefaults };
        let schema = { ...globalSchema };

        // syncronize cursor of the schema & json to "webrtc" root
        schema = schema.properties[NS].properties;
        defaults = defaults[NS];

        // traverse the schema to create a flattened array of metadata
        const array = traverse(schema, defaults);
        array.forEach(row => {
            const [path, key, value, validate, type, title = "", description] = row;
            const props = { title, description, type, validate, path };
            const id = [...path, key].join("-");

            // An "object" only contains meta data (no form element required)
            if (type === "object" && !description && !title) {
                return renderSeperator(key, path);
            }

            // Alway have a title
            props.title = props.title || key;

            // keep track of data points ONLY
            if (!/^(array|object)$/.test(type)) {
                registry[id] = row;
            }

            renderFormRow(props, value, id);
        });
    }

    function renderFormRow(props, value, id) {
        // returns undefined
        //
        // Inserts or Updates the form row
        // by appending a new DOMHtmlDiv or
        // by updating HTMLInputElement attributes

        const row = getHTMLDivRow(props, value, id);
        const prev = document.querySelector(`#${id}`);
        let deltas = {};

        if (!row) return null;

        if (!prev) {
            container.append(row);
            return;
        }

        // Element could be: single input or radio group
        const inputs = row.querySelectorAll(".op-value");

        // Preserve existing DOM element and update properties
        prev.querySelectorAll(".op-value").forEach((item, n) => {
            const input = inputs[n];

            // managed storage is read-only
            if (globalReadOnly) {
                item.setAttribute("disabled", "disabled");
            }

            // update with new value
            if (item.value !== input.value) {
                deltas.value = true;
                item.value = input.value;
                item.classList.add("op-changed");

                // remove highlight
                setTimeout(() => {
                    // item.classList.remove("op-changed")
                    item.classList.remove("op-changed");
                }, 2500);
            }

            // exit if this is a single input (ie. not radio group)
            if (!(inputs.length - 1)) return;

            // reset state of radio and select option
            ["checked"].forEach(attr => {
                const bothHave = input.hasAttribute(attr) && item.hasAttribute(attr);
                const bothDont = !input.hasAttribute(attr) && !item.hasAttribute(attr);

                if (bothHave || bothDont) return;

                item.removeAttribute(attr);
                if (input.hasAttribute(attr)) item.setAttribute(attr, attr);
                deltas[attr] = true;
            });
        });

        // report changed properties
        deltas = Object.keys(deltas);
        if (deltas.length) {
            console.log("Element changed (%s)", id, deltas);
        }
    }

    function traverse(schema, json, path = [], history = []) {
        // returns Array of metadata parts
        //
        // recursively tarverse the schema and create the metadata
        // metadata is derived from managed-schema.json

        Object.keys(schema).forEach(key => {
            const {
                title = null,
                description = null,
                type,
                properties = null,
                items = null,
                validate = null
            } = schema[key];
            const validator = validate || type;
            const primitive = !/(object|array)/i.test(type);
            const value = json[key];
            let spread;

            if (/^(object)$/i.test(type)) {
                history.push([path, key, null, null, type, title, description]);
                traverse(properties, value, [...path, key], history);
                return;
            }

            if (/^(array)$/i.test(type)) {
                history.push([[...path], key, null, null, type, title, description]);
                spread = value.map(() => items);
                traverse(spread, value, [...path, key], history);
                return;
            }

            if (primitive) {
                history.push([path, key, value, validator, type, title, description]);
            }
        });
        return history;
    }

    function getHTMLDivRow(props, value, id) {
        // returns HTMLDivElement w. event bindings on inputs
        //
        // A row Object is a DOM element that contains
        // addtional meta data to describe the input
        // with generic event bindings that validate the data

        const row = document.createElement("div");
        const { title, description = null, type, validate = null, path = [] } = props;

        const className = path && path.length && path[0];
        const input = getFormField(type, value, id, validate);
        const label = `<span class="label" ${i18n(title)}></span>`;
        const wording = description ? `<div class="wording" ${i18n(description)}></div>` : "";

        row.id = id;
        row.classList.add("opItem");
        row.classList.add(`op-${type.toLowerCase()}`);
        className ? row.classList.add("op-group") : null;
        className ? row.classList.add(className) : null;
        row.dataset.type = type;
        row.innerHTML = `${label} ${input} ${wording}`;

        if (/type="text"/i.test(input)) {
            row.querySelector("input").onchange = updateTextValue;
        }

        if (/type="radio"/gi.test(input)) {
            row.querySelectorAll("input[type=radio]").forEach(
                el => (el.onclick = updateRadioValue)
            );
        }

        return row;
    }

    function i18n(string, type = "content") {
        // returns String used as HTML data attribute
        return `data-i18n-${type}="${t(string)}"`;
    }

    function t(string) {
        // returns String
        //
        // formatted as uppercase sanitized string for i18n translation
        // removes punctuation and parenthesis
        return `${string}`
            .trim()
            .toUpperCase()
            .replace(/[,.()'"`:;]+/g, " ")
            .replace(/\s+/g, "_")
            .trim();
    }

    function getFormField(type, value = "", id, validate) {
        // returns String (aka UI input element innerHTML)
        //
        // Uses json-schema definition to render metadata
        // appends HTML arrtibutes for i18 translation parser

        let html, truthy;

        const capitalize = type.replace(/^(.)/, c => c.toUpperCase());
        const placeholder = vaildateFormField(validate) !== null ? "Optional" : capitalize;
        const disabled = globalReadOnly ? `disabled="disabled"` : "";

        switch (type) {
            case "string":
            case "number":
            case "integer":
                html = `
                        <input type="text"
                            id="${id}"
                            class="op-value"
                            value="${value}"
                            ${i18n(placeholder, "placeholder")}
                            ${disabled}
                            placeholder="${placeholder}"
                            data-validate="${validate}">`;
                break;

            case "boolean":
                truthy = /^(true|yes)$/.test(value);
                html = `
                        <input type="radio"
                            id="${id}-true"
                            name="${id}"
                            class="op-value"
                            value="true"
                            ${disabled}
                            ${truthy ? "checked=''" : ""}>
                        <label for="${id}-true" ${i18n("yes")}></label>
                        <input type="radio"
                            id="${id}-false"
                            name="${id}"
                            class="op-value"
                            value="false"
                            ${disabled}
                            ${!truthy ? "checked=''" : ""}>
                        <label for="${id}-false" ${i18n("no")}></label>`;
                break;

            case "array":
            case "object":
                html = `<!-- ${type} ${id} -->`;
                break;

            default:
                html = `<!-- ${type} ${value} ${id} -->`;
                console.log("WARNING!! Form field dropthrough", type, value, id);
                break;
        }
        return html;
    }

    function updateRadioValue() {
        let { id, value } = this;

        // remove boolean flag from ID
        id = id.replace(/-(true|false)$/, "");

        // cast value as Boolean
        value = /^true$/i.test(value);

        persistToLocalStorage(id, value);
    }

    function updateTextValue() {
        const { id, value } = this;
        const validator = this.dataset.validate;
        const isvalid = vaildateFormField(validator, value);
        this.style.background = "";

        if (isvalid === null) {
            this.style.background = "rgba(255,0,0,0.3)";
            return;
        }
        this.value = isvalid;

        persistToLocalStorage(id, value);
    }

    function persistToLocalStorage(id, value) {
        if (globalReadOnly) {
            console.log("WARNING! read-only");
            return;
        }

        const paths = id.split("-");
        const webrtc = setValue({ ...globalDefaults[NS] }, paths, value);

        chrome.storage.local.set({ [NS]: webrtc }, () => {
            console.log("Persist complete", chrome.runtime.lastError);
            globalDefaults[NS] = webrtc;
        });
    }

    function vaildateFormField(validator, value = null) {
        // returns String (pass) or null (failed)
        // @validator can be declared datatype OR Regexp String

        value = `${value}`.trim().replace(/^(null)$/i, "");
        const rex = /^\/(.*)\/([gim]+)?$/;
        const isregex = rex.test(validator);

        const datatypes = {
            string: /^.+$/i,
            number: /^\d+(\.\d+)?$/,
            integer: /^\d+$/,
            ipaddress: /^(localhost|\d{2,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/,
            boolean: /^(true|false)$/i
        };

        let re = datatypes[validator] || datatypes.string;

        if (isregex) {
            const [_, pattern, mod = ""] = validator.match(rex);
            re = new RegExp(pattern, mod);
        }
        return re && re.test(value) ? value : null;
    }

    function setValue(obj, path, value) {
        if (Object(obj) !== obj) return obj; // When obj is not an object
        // If not yet an array, get the keys from the string-path
        if (!Array.isArray(path)) path = path.toString().match(/[^.[\]]+/g) || [];
        path.slice(0, -1).reduce(
            (a, c, i) =>
                Object(a[c]) === a[c] // Does the key exist and is its value an object?
                    ? // Yes: then follow that path
                      a[c]
                    : // No: create the key. Is the next key a potential array-index?
                      (a[c] =
                          Math.abs(path[i + 1]) >> 0 === +path[i + 1]
                              ? [] // Yes: assign a new array object
                              : {}), // No: assign a new plain object
            obj
        )[path.pop()] = value; // Finally assign the value to the last key
        return obj; // Return the top-level object to allow chaining
    }
});
